"""
    lmm_woodbury(y, X, Z)

Fitting an linear mixed model with responses `y`, fixed effects `X`, and random 
effects `Z[1]`, ..., `Z[end]`, assuming `Y` is normal with mean `X * β`, 
and covariance `σ2[1]*Z[1]*Z[1]'+...+σ2[end-1]*Z[end]*Z[end]+σ2[end]*I`. 

This method implements MM algorithm using the Woodbury formula, and is 
applicable to problems where the total number of random effects 
`sum(M -> size(M, 2), Z)` is much smaller than the number of observations.

# Arguments
- `y`: n-by-1 response vector.  
- `X`: n-by-p predictor matrix.  
- `Z`: `Z[i]` is the n-by-qi predictor matrix for the i-th random effects.  

# Keyword Arguments
- `verbose`: logical, verbose display.
- `maxiter`: maximum number of iterations.
- `funtol`: covergence tolerance for objective value.
- `β`: starting value of fixed effects.
- `σ2`: staring value of variance components.
- `algo`: `:MM` (default) or `:EM` of `:FS`.

# Output
- `β`: fitted fixed effects.
- `σ2`: fitted random effects.
- `nter`: number of iterations.
- `logl`: fitted log-likelihood.
"""
function lmm_woodbury(
    y::AbstractVector{T}, 
    X::AbstractMatrix{T},
    Z::Vector{<:AbstractMatrix{T}};
    verbose::Bool = false, 
    maxiter::Integer = 5000, 
    funtol::Number = 1e-8,
    β::Vector{T} = X \ y,
    σ2::Vector{T} = ones(T, length(Z) + 1),
    algo::Symbol = :MM
    ) where T <: LinAlg.BlasFloat
    
    # dimensions
    n, p = size(X)
    m    = length(Z) # no. non-identity variance components  
    qi   = [size(Z[i], 2) for i in 1:m]
    qidx = [[1:qi[1]]; [(sum(qi[1:i-1])+1):sum(qi[1:i]) for i in 2:m]]
    q    = sum(qi)
    # pre-compute fixed quantities
    ztx = [Z[i]' * X    for i in 1:m]           # ztx[i] = Z[i]' * X
    zty = [Z[i]' * y    for i in 1:m]
    ztz = [Z[i]' * Z[j] for i in 1:m, j in 1:m] # ztz[i, j] = Z[i]'Z[j]
    zi2 = [vecnorm(Z[i])^2 for i in 1:m]
    xtx = X'X
    xty = X'y
    # pre-allocate working arrays
    if algo == :EM
        zrk = [rank(Z[i]) for i in 1:m]
    elseif algo == :FS
        z     = log.(σ2) # z is log transformed variables
        znew  = similar(z)
        Δz    = similar(z)
        ztoiz = deepcopy(ztz)
        σnew  = similar(σ2)
        gradz = zeros(T, m + 1)
        infoz = zeros(T, m + 1, m + 1)
        storagez = deepcopy(Z)
    end
    storagen  = zeros(T, n)    # n-by-1 vector
    σztr      = zeros(T, q)    # q-by-1 vector
    storageq  = zeros(T, q)    # q-by-1 vector
    storagep  = zeros(T, p)    # p-by-1 vector
    M         = zeros(T, q, q) # q-by-q matrix
    storageqq = zeros(T, q, q) # q-by-q matrix
    storageqp = zeros(T, q, p) # q-by-p matrix
    storagepp = zeros(T, p, p) # p-by-p matrix
    # pre-allocate views/pointers
    storageqi   = [view(storageq , qidx[i])    for i in 1:m]
    storageqpi  = [view(storageqp, qidx[i], :) for i in 1:m]
    storageqqi  = [view(storageqq, qidx[i], :) for i in 1:m]
    storageqqj  = [view(storageqq, :, qidx[j]) for j in 1:m]
    storageqqij = [view(storageqq, qidx[i], qidx[j]) for i in 1:m, j in 1:m]
    # calculate initial objective value
    σ = sqrt.(σ2)
    res = y - X * β
    update_σztr!(σztr, Z, res, σ, qidx)
    update_woodbury!(storageqp, storageq, storageqq, M, ztx, zty, ztz, σ, qidx)
    Mchol = cholfact!(M, :L)
    A_ldiv_B!(Mchol[:L], σztr)
    loglconst = - n * log(2π)
    logl = (loglconst - - (n - q) * log(σ2[end]) - logdet(Mchol) - 
        (vecnorm(res)^2 - vecnorm(σztr)^2) / σ2[end]) / 2
    verbose && println("iter=", 0, " logl=", logl)
    verbose && println("β=", β," σ2=", σ2)
    # MM/EM/FS loop
    niter = maxiter
    for iter in 1:maxiter
        # storagen = Ω^{-1} * res
        copy!(storageq, σztr)
        At_ldiv_B!(Mchol[:L], storageq) # storageq = inv(L) * σztr
        copy!(storagen, res)
        for i in 1:m
            BLAS.gemv!('N', -σ[i], Z[i], storageqi[i], T(1), storagen)
        end
        # storageqq = inv(L) * diag(σ) * Z' * Z
        A_ldiv_B!(Mchol[:L], storageqq)
        if algo == :MM || algo == :EM
            # update variance components σ1,..., σm
            for i in 1:m
                At_mul_B!(storageqi[i], Z[i], storagen)
                qf = (vecnorm(storageqi[i]) / σ2[end])^2 # quadratic form
                tr = (zi2[i] - vecnorm(storageqqj[i])^2) / σ2[end] # trace term
                if algo == :MM
                    σ2[i] *= sqrt(qf / tr)
                elseif algo == :EM
                    σ2[i] += (qf - tr) / zrk[i] * σ2[i]^2
                end
            end
            # update σ0
            qf = (vecnorm(storagen) / σ2[end])^2
            At_ldiv_B!(Mchol[:L], storageqq) # storageqq is inv(M) * diag(σ) * Z' * Z
            tr = T(n)
            for i in 1:m
                for j in qidx[i]
                    tr -= σ[i] * storageqq[j, j]
                end
            end
            tr /= σ2[end]
            if algo == :MM
                σ2[end] *= sqrt(qf / tr)
            elseif algo == :EM
                σ2[end] += (qf - tr) / n * σ2[end]^2
            end
            σ .= sqrt.(σ2)
            # update M and other quantities
            update_woodbury!(storageqp, storageq, storageqq, M, ztx, zty, ztz, σ, qidx)
            Mchol = cholfact!(M, :L)
        elseif algo == :FS
            # gradz and infoz
            for j in 1:m
                At_mul_B!(storageqi[j], Z[j], storagen)
                qf = (vecnorm(storageqi[j]) / σ2[end])^2 # quadratic form
                tr = (zi2[j] - vecnorm(storageqqj[j])^2) / σ2[end] # trace term
                gradz[j] = σ2[j] * (qf - tr) / 2
                for i in j:m
                    copy!(ztoiz[i, j], ztz[i, j])
                    BLAS.gemm!('T', 'N', T(-1), storageqqj[i], storageqqj[j], T(1), ztoiz[i, j])
                    infoz[i, j] = σ2[i] * σ2[j] * vecnorm(ztoiz[i, j])^2 / 2 / σ2[end]^2
                    infoz[j, i] = infoz[i, j]
                end
            end
            # last entry of gradz
            qf = (vecnorm(storagen) / σ2[end])^2
            At_ldiv_B!(Mchol[:L], storageqq) # storageqq is inv(M) * diag(σ) * Z' * Z
            tr = T(n)
            for i in 1:m, j in qidx[i]
                tr -= σ[i] * storageqq[j, j]
            end
            tr /= σ2[end]
            gradz[m + 1] = σ2[end] * (qf - tr) / 2
            # last column/row of infoz
            for j in 1:m
                copy!(storagez[j], Z[j])
                for i in 1:m
                    BLAS.gemm!('N', 'N', -σ[i], Z[i], storageqqij[i, j], T(1), storagez[j])
                end
                infoz[m + 1, j] = σ2[j] * vecnorm(storagez[j])^2 / σ2[end] / 2
                infoz[j, m + 1] = infoz[m + 1, j]
            end
            infoz[m + 1, m + 1] = T(n)
            for j in 1:m, i in qidx[j]
                infoz[m + 1, m + 1] -= 2 * σ[j] * storageqq[i, i]
                for k in 1:m, l in qidx[k]
                    infoz[m + 1, m + 1] += σ[j] * σ[k] * storageqq[i, l] * storageqq[l, i]
                end
            end
            infoz[m + 1, m + 1] /= 2
            # make sure the information matrix is positive definite
            if eigmin(infoz) ≤ 1e-3; infoz += T(1e-3)I; end
            # line search
            A_ldiv_B!(Δz, cholfact!(infoz), gradz) # Newton direction
            stepsize = T(1)
            for btiter in 1:50
                znew .= z .+ stepsize .* Δz
                σnew .= exp.(znew ./ 2)
                update_woodbury!(storageqp, storageq, storageqq, M, ztx, zty, ztz, σnew, qidx)
                if eigmin(M) > 0
                    Mchol = cholfact!(M, :L)
                else
                    stepsize /= 2
                    continue
                end
                update_σztr!(σztr, Z, res, σnew, qidx)
                A_ldiv_B!(Mchol[:L], σztr)
                loglnew = (loglconst - 2(n - q) * log(σnew[end]) - logdet(Mchol) - 
                    (vecnorm(res)^2 - vecnorm(σztr)^2) / σnew[end]^2) / 2
                if loglnew > logl
                    break
                elseif btiter == 50
                    warn("line search failed")
                else
                    stepsize /= 2
                end
            end  # end of line search
            copy!(z, znew)
            copy!(σ, σnew)
            σ2 .= σ .* σ
        end
        # update fixed effects
        A_ldiv_B!(Mchol[:L], storageqp) # storageqp is inv(L) * σztx 
        At_mul_B!(storagepp, storageqp, storageqp) 
        storagepp .= xtx .- storagepp # Gram matrix
        if !isposdef(storagepp)
            storagepp += 1e-3I
        end
        A_ldiv_B!(Mchol[:L], storageq) # storageq is inv(L) * σzty 
        copy!(β, xty)
        BLAS.gemv!('T', T(-1), storageqp, storageq, T(1), β) # rhs
        A_ldiv_B!(cholfact!(Symmetric(storagepp)), β)
        # update residuals
        copy!(res, y)
        BLAS.gemv!('N', T(-1), X, β, T(1), res)
        update_σztr!(σztr, Z, res, σ, qidx)
        A_ldiv_B!(Mchol[:L], σztr)
        # check convergence
        loglold = logl
        logl = (loglconst - (n - q) * log(σ2[end]) - logdet(Mchol) - 
            (vecnorm(res)^2 - vecnorm(σztr)^2) / σ2[end]) / 2
        verbose && println("iter=", iter, " logl=", logl)
        verbose && println("β=", β," σ2=", σ2)
        if abs(logl - loglold) < funtol * (abs(logl) + 1)
            niter = iter
            break
        end
    end
    # output
    return β, σ2, niter, logl
end

"""
    update_woodbury!()

Update quantities in Woodbury formula.
"""
function update_woodbury!(
    σztx::Matrix{T},
    σzty::Vector{T},
    σztz::Matrix{T},
    M::Matrix{T},
    ztx::Vector{<:AbstractMatrix{T}},
    zty::Vector{<:AbstractVector{T}},
    ztz::Matrix{<:AbstractMatrix{T}},
    σ::Vector{T},
    qidx::Vector{<:UnitRange}
    ) where T <: AbstractFloat
    m = length(qidx)
    for j in 1:m
        σztx[qidx[j], :] = σ[j] * ztx[j]
        σzty[qidx[j]] = σ[j] * zty[j]
        for i in 1:m
            σztz[qidx[i], qidx[j]] = σ[i] * ztz[i, j]
            M[qidx[i], qidx[j]]    = σ[i] * σ[j] * ztz[i, j]
        end
    end
    σ02 = σ[end] * σ[end]
    for i in 1:size(M, 1)
        M[i, i] += σ02
    end
    σztx, σzty, σztz, M
end

function update_σztr!(
    σztr::Vector{T},
    Z::Vector{<:AbstractMatrix{T}},
    r::Vector{T},
    σ::Vector{T},
    qidx::Vector{<:UnitRange}
    ) where T <: AbstractFloat
    m = length(qidx)
    for i in 1:m
        @views At_mul_B!(σztr[qidx[i]], Z[i], r)
        σztr[qidx[i]] .*= σ[i]
    end
    σztr
end

