#
# This simulation comapre the performance among MM, EM, FS, acceleratated MM
# and lme4(relMM wrapper)
# comparison metrics include:
#     1. Runtime
#     2. iterations
#     3. objective
#     4. MSE of βhat
#     5. MSE of σ2hat
#
# Data: Mendel 29 data
# 2 variance components: kinship(GRM), environmental error
#
include("../vc.jl")
include("../vc2.jl")
include("../lmm.jl")
using RCall
R"""
library(MixedModels)
library(pedigreemm)
"""

srand(123)

y = vec(readdlm("y.txt"))
n = length(y)
gender = readdlm("gender.txt")
X = hcat(ones(n), gender)
famid = readdlm("famid.txt")
kinship = readdlm("kinship.txt")
kinship_MoM = readdlm("MoM.txt", ',')
kinship_GRM = readdlm("GRM.txt", ',')

# two variance components: additive and environment
V2 = Array{Matrix{Float64}}(2)
V2[1] = kinship_GRM
V2[2] = eye(n)
M2 = Array{Matrix{Float64}}(2)
M2[1] = cholfact(kinship_GRM)[:L]
M2[2] = eye(n)

# simulation
V = V2
M = M2
m = length(V)
p = size(X, 2)
β = [33, 6.0] # coef of sex effect
σ2 = [1.0 , 2.0] # variance of kinsip(additive) and error(environment)
vcratio = 1
σ2[1] = σ2[end] * vcratio
Ω = zeros(n, n)
for i = 1:m
    Ω += σ2[i] * V[i]
end
y = X * β + cholfact(Symmetric(Ω))[:L] * randn(n)

vc(y, X, V; algo=:MM, maxiter=5) # trigger compilation
vc(y, X, V[1]; algo=:MM, maxiter=5) # trigger compilation

println("MM algorithm using eigen-decomposition")
@time βhat, σ2hat, iter, obj = vc(y, X, V[1]; algo=:MM)
@show βhat, σ2hat, iter, obj
println()

println("MM algorithm")
@time βhat, σ2hat, iter, obj = vc(y, X, V; algo=:MM)
@show βhat, σ2hat, iter, obj
println()

println("EM algorithm using eigen-decomposition")
@time βhat, σ2hat, iter, obj = vc(y, X, V[1]; algo=:EM)
@show βhat, σ2hat, iter, obj
println()

println("EM algorithm")
@time βhat, σ2hat, iter, obj = vc(y, X, V; algo=:EM)
@show βhat, σ2hat, iter, obj
println()

println("FS algorithm using eigen-decomposition")
@time βhat, σ2hat, iter, obj = vc(y, X, V[1]; algo=:FS)
@show βhat, σ2hat, iter, obj
println()

println("FS algorithm")
@time βhat, σ2hat, iter, obj = vc(y, X, V; algo=:FS)
@show βhat, σ2hat, iter, obj
println()

# Call mm in R package MixedModels, a wrapper of lme4
println("lme4")
A1 = V[1]
@rput A1
@rput n
R"""
df <- data.frame(y=$y, X = $X)
rownames(A1) <- as.character(seq(1, n))
g1 <- seq(1, n)
ptm <- proc.time()
model <- mm(formula = y ~ 0 + X.1 + X.2 + (1 | g1), covarrel = list(g1 = A1),
    data = df, REML = F,
    control = lmerControl(check.nobs.vs.nlev="ignore",
    check.nobs.vs.nRE="ignore"))
time_lme4 <- proc.time() - ptm
elapsed_lme4 <- time_lme4[3]
beta_lme4 <- fixef(model)
vc_lme4 <- VarCorr(model)
vc_lme4_res <- attr(vc_lme4, "sc")
logL_lme4 <- as.numeric(logLik(model))
"""
@rget beta_lme4
@rget vc_lme4
@rget elapsed_lme4
@rget logL_lme4
@rget vc_lme4_res
βhat  = beta_lme4
σ2hat = [vc_lme4[:g1], vc_lme4_res]
t = elapsed_lme4
obj = logL_lme4
@show t
@show βhat, σ2hat, obj
