#
# This simulation comapre the performance among MM, EM, FS, acceleratated MM
# and lme4(relMM wrapper)
# comparison metrics include:
#     1. Runtime
#     2. iterations
#     3. objective
#     4. MSE of βhat
#     5. MSE of σ2hat
#
# Data: Mendel 29 data
# 2 variance components: kinship(GRM), environmental error
#
include("../vc.jl")
include("../vc2.jl")
include("../lmm.jl")
using JLD
using RCall
R"""
library(MixedModels)
library(pedigreemm)
"""

y = vec(readdlm("y.txt"))
n = length(y)
gender = readdlm("gender.txt")
X = hcat(ones(n), gender)
famid = readdlm("famid.txt")
kinship = readdlm("kinship.txt")
kinship_MoM = readdlm("MoM.txt", ',')
kinship_GRM = readdlm("GRM.txt", ',')

# two variance components: additive and environment
V2 = Array{Matrix{Float64}}(2)
V2[1] = kinship_GRM
V2[2] = eye(n)
M2 = Array{Matrix{Float64}}(2)
M2[1] = cholfact(kinship_GRM)[:L]
M2[2] = eye(n)

# simulation under different settings
V = V2
M = M2
m = length(V)
p = size(X, 2)
β = [33, 6.0] # coef of sex effect

if m == 4
    σ2 = [1.0 , 2.0, 3.0, 2]  # variance of kinsip(additive), delta7(dominance),
    # H(householder) and error(environment)
elseif m == 2
    σ2 = [1.0 , 2.0] # variance of kinsip(additive) and error(environment)
end

Ω = zeros(n, n)
nmethod = 4
nsim = 50
vcratio = [0 0.05 0.1 1 10 20]
nratio = length(vcratio)
mendelresult = zeros(6, nmethod, nratio)
mse_β = zeros(2, nmethod, p, nratio)
mse_σ2 = zeros(2, nmethod, m, nratio)
obj = zeros(nsim, nmethod)
t = zeros(nsim, nmethod)
iter = ones(Int, nsim, nmethod)
σ2sqerr = zeros(nmethod, m, nsim)
βsqerr = zeros(nmethod, p, nsim)

for j in 1:nratio
    println("vcratio = ", vcratio[j])
    for k in 1:nsim
        σ2[1] = σ2[end] * vcratio[j]
        Ω = zeros(Ω)
        for i = 1:m
            Ω += σ2[i] * V[i]
        end
        srand(13j  + 7k)
        y = X * β + cholfact(Symmetric(Ω))[:L] * randn(n)

        t[k, 1] = @elapsed βhat, σ2hat, iter[k, 1], obj[k, 1] = vc(y, X, V[1]; algo=:MM)
        σ2sqerr[1, :, k] = sqrt.((σ2hat - σ2).^2)
        βsqerr[1, :, k ] = sqrt.((βhat - β).^2)
        
        t[k, 2] = @elapsed βhat, σ2hat, iter[k, 2], obj[k, 2] = vc(y, X, V[1]; algo=:EM)
        σ2sqerr[2, :, k] = sqrt.((σ2hat - σ2).^2);
        βsqerr[2, :, k] = sqrt.((βhat - β).^2);
        
        t[k, 3] = @elapsed βhat, σ2hat, iter[k, 3], obj[k, 3] = vc(y, X, V[1]; algo=:FS)
        σ2sqerr[3, :, k] = sqrt.((σ2hat - σ2).^2)
        βsqerr[3, :, k] = sqrt.((βhat - β).^2)

        # Call mm in R package MixedModels, a wrapper of lme4
        A1 = V[1]
        @rput A1
        @rput n
        R"""
        df <- data.frame(y=$y, X = $X)
        rownames(A1) <- as.character(seq(1, n))
        g1 <- seq(1, n)
        ptm <- proc.time()
        model <- mm(formula = y ~ 0 + X.1 + X.2 + (1 | g1), covarrel = list(g1 = A1),
            data = df, REML = F,
            control = lmerControl(check.nobs.vs.nlev="ignore",
            check.nobs.vs.nRE="ignore"))
        time_lme4 <- proc.time() - ptm
        elapsed_lme4 <- time_lme4[3]
        beta_lme4 <- fixef(model)
        vc_lme4 <- VarCorr(model)
        vc_lme4_res <- attr(vc_lme4, "sc")
        logL_lme4 <- as.numeric(logLik(model))
        """
        @rget beta_lme4
        @rget vc_lme4
        @rget elapsed_lme4
        @rget logL_lme4
        @rget vc_lme4_res
        βhat  = beta_lme4
        σ2hat = [vc_lme4[:g1], vc_lme4_res]
        t[k, 4] = elapsed_lme4
        obj[k, 4] = logL_lme4
        # RMSE
        σ2sqerr[4, :, k] = sqrt.((σ2hat - σ2).^2)
        βsqerr[4, :, k] = sqrt.((βhat - β).^2)
        
    end  # end of loop over Nsim

    mendelresult[1, :, j] = mean(iter, 1)
    mendelresult[2, :, j] = std(iter, 1)
    t = t .* 1e3
    mendelresult[3, :, j] = mean(t, 1)
    mendelresult[4, :, j] = std(t, 1)
    mendelresult[5, :, j] = mean(obj, 1)
    mendelresult[6, :, j] = std(obj, 1)
    mse_β[1, :, :, j] = mean(βsqerr, 3)
    mse_β[2, :, :, j] = std(βsqerr, 3)
    mse_σ2[1, :, :, j] = mean(σ2sqerr, 3)
    mse_σ2[2, :, :, j] = std(σ2sqerr, 3)
    
end # end of loop over ratio

# machine information
juliainfo = IOBuffer()
versioninfo(juliainfo)
juliainfo = String(take!(juliainfo))
print(juliainfo)
R"""
rinfo = sessionInfo()
"""
@rget rinfo

# save workspace
@save "mendel29_sim_result.jld"
