"""
    vc(y, X, V)

Fitting a variance component model with responses `y`, fixed effects `X`, and 
random effects `Z[1]`, ..., `Z[end]`, assuming `Y` is normal with mean 
`X * β`, and covariance `σ2[1]*V[1]+...+σ2[end]*V[end]`. 

# Arguments
- `y`: n-by-1 response vector.  
- `X`: n-by-p predictor matrix.  
- `V`: `V[i]` is the i-th psd matrix.  

# Keyword Arguments
- `verbose`: logical, verbose display.
- `maxiter`: maximum number of iterations.
- `funtol`: covergence tolerance for objective value.
- `β`: starting value of fixed effects.
- `σ2`: staring value of variance components.
- `algo`: `:MM` (default) or `:EM` or `:FS`

# Output
- `β`: fitted fixed effects.
- `σ2`: fitted random effects.
- `nter`: number of iterations.
- `logl`: fitted log-likelihood.
"""
function vc(
    y::AbstractVector{T}, 
    X::AbstractMatrix{T},
    V::Vector{<:AbstractMatrix{T}};
    verbose::Bool = false, 
    maxiter::Integer = 5000, 
    funtol::Number = 1e-8,
    β::Vector{T} = X \ y,
    σ2::Vector{T} = ones(T, length(V)),
    algo::Symbol = :MM
    ) where T <: LinAlg.BlasFloat

    n, p = size(X)
    m    = length(V)   # no. variance components
    # pre-allocate working arrays
    storagen  = zeros(T, n)
    storagenp = zeros(T, n, p)
    storagepp = zeros(T, p, p)
    storagep  = zeros(T, p)
    if algo == :EM
        zrk = [rank(V[i]) for i in 1:m]
    elseif algo == :FS
        z     = log.(σ2) # z is log transformed variables
        znew  = similar(z)
        σ2new = similar(σ2)
        gradz = zeros(T, m)
        infoz = zeros(T, m, m)
        Δz    = zeros(T, m)    
    end
    # initialize algorithm
    Ω = zeros(T, n, n)
    update_Ω!(Ω, σ2, V)
    Ωchol = cholfact!(Ω, :L)
    Ωinv  = inv(Ωchol)
    res   = y - X * β
    copy!(storagen, res)
    A_ldiv_B!(Ωchol[:L], storagen)
    loglConst = - n * log(2π)
    logl = (loglConst - logdet(Ωchol) - vecnorm(storagen)^2) / 2
    verbose && println("iter=", 0," logl=", logl)
    verbose && println("β=", β," σ2=", σ2)
    niter = maxiter
    # MM/EM/FS loop
    for iter in 1:maxiter
        # update variance components
        At_ldiv_B!(Ωchol[:L], storagen) # stroagen is inv(Ω) * residual
        if algo == :MM || algo == :EM
            for j in 1:m
                qf = dot(V[j] * storagen, storagen)
                tr = dot(Ωinv, V[j])
                if algo == :MM
                    σ2[j] *= sqrt(qf / tr)
                elseif algo == :EM
                    σ2[j] += (qf - tr) / zrk[j] * σ2[j]^2
                end
            end
            # update Ω
            update_Ω!(Ω, σ2, V)
            Ωchol = cholfact!(Ω, :L)
        elseif algo == :FS
            for j in 1:m
                gradz[j] = σ2[j] * (dot(V[j] * storagen, storagen) - dot(Ωinv, V[j])) / 2
                for k in j:m
                    infoz[k, j] = σ2[j] * σ2[k] * trace(Ωinv * V[j] * Ωinv * V[k]) / 2
                    infoz[j, k] = infoz[k, j]
                end
            end
            # make sure the information matrix is positive definite
            if eigmin(infoz) ≤ 1e-3; infoz += T(1e-3)I; end
            # line search
            A_ldiv_B!(Δz, cholfact!(infoz), gradz) # Newton direction
            stepsize = T(1)
            for btiter in 1:50
                znew  .= z .+ stepsize .* Δz
                σ2new .= exp.(znew)
                update_Ω!(Ω, σ2new, V)
                if eigmin(Ω) > 0
                    Ωchol = cholfact!(Ω, :L)
                else
                    stepsize /= 2
                    continue
                end
                A_ldiv_B!(storagen, Ωchol, res)
                loglnew = (loglConst - logdet(Ωchol) - dot(res, storagen)) / 2
                if loglnew > logl
                    break
                elseif btiter == 50
                    warn("line search failed")
                else
                    stepsize /= 2
                end                
            end  # end of line search
            copy!(z, znew)
            copy!(σ2, σ2new)    
        end
        Ωinv = inv(Ωchol)
        # update fixed effects
        A_mul_B!(storagenp, Ωinv, X)
        At_mul_B!(storagepp, X, storagenp)
        A_mul_B!(storagen, Ωinv, y)
        At_mul_B!(storagep, X, storagen)
        copy!(β, storagep)
        if !isposdef(storagepp)
            storagepp += T(1e-3)I
        end
        A_ldiv_B!(cholfact!(Symmetric(storagepp)), β)
        A_mul_B!(res, X, β)
        res .= y .- res
        copy!(storagen, res)
        A_ldiv_B!(Ωchol[:L], storagen)
        # check convergence
        loglold = logl
        logl = (loglConst - logdet(Ωchol) - vecnorm(storagen)^2) / 2
        verbose && println("iter=", iter," logl=", logl)
        verbose && println("β=", β," σ2=", σ2)
        if abs(logl - loglold) < funtol * (abs(logl) + 1)
            niter = iter
            break
        end
    end
    # output
    return β, σ2, niter, logl
end

function update_Ω!(
    Ω::AbstractMatrix{T},
    σ2::AbstractVector{T},
    V::Vector{<:AbstractMatrix{T}}
    ) where T <: AbstractFloat
    fill!(Ω, 0)
    for j in 1:length(σ2)
        Ω .+= σ2[j] .* V[j]
    end
    Ω
end

function update_Ω!(
    Ω::AbstractMatrix{T},
    σ2::AbstractVector{T},
    V::Vector{<:AbstractMatrix{T}}
    ) where T <: LinAlg.BlasFloat
    fill!(Ω, 0)
    for j in 1:length(σ2)
        BLAS.axpy!(σ2[j], V[j], Ω)
    end
    Ω
end
