"""
    vc(y, X, V1, V2)

Fitting a variance component model with r̃ponses `y`, fixed effects `X`, and 
variance component matrices `V1` and `V2`, assuming `Y` is normal with mean 
`X * β`, and covariance `σ2[1]*V1++σ2[2]*V2`.

This method implements the special algorithm using eigendecomposition of `V1`.

# Arguments
- `y`: n-by-1 r̃ponse vector.  
- `X`: n-by-p predictor matrix.  
- `V1`: n-by-n positive semidefinite matrix.  
- `V2`: n-by-n positive definite matrix.

# Keyword Arguments
- `verbose`: logical, verbose display.
- `maxiter`: maximum number of iterations.
- `funtol`: covergence tolerance for objective value.
- `β`: starting value of fixed effects.
- `σ2`: staring value of variance components.
- `algo`: `:MM` (default) or `:EM` or `:FS`

# Output
- `β`: fitted fixed effects.
- `σ2`: fitted random effects.
- `nter`: number of iterations.
- `logl`: fitted log-likelihood.
"""
function vc(
    y::AbstractVector{T}, 
    X::AbstractMatrix{T},
    V1::AbstractMatrix{T},
    V2::Union{AbstractMatrix{T}, LinAlg.UniformScaling} = I;
    verbose::Bool = false, 
    maxiter::Integer = 5000, 
    funtol::Number = 1e-8,
    β::Vector{T} = X \ y,
    σ2::Vector{T} = ones(T, 2),
    algo::Symbol = :MM
    ) where T <: LinAlg.BlasFloat

    n, p = size(X)
    # pre-compute eigen-decomposition
    if V2 == I
        F = eigfact(Symmetric(V1))
    else
        F = eigfact(Symmetric(V1), Symmetric(V2)) # F[:values], F[:vectors]
    end
    d = map(di -> max(di, 0), F[:values]) # make sure eigenvalues are nonnegative
    Φ = F[:vectors]
    # pre-allocate working arrays
    ỹ = Φ'y
    X̃ = Φ'X
    r̃ = ỹ - X̃ * β
    wt  = σ2[1] * d + σ2[2]
    storagen  = r̃ .* r̃ ./ wt
    storagenp = zeros(T, n, p)
    storagepp = zeros(T, p, p)
    storagep  = zeros(T, p)
    if algo == :EM
        rk1 = V2 == I ? n : rank(V1)
    elseif algo == :FS
        z     = log.(σ2)
        znew  = similar(z)
        wtnew = similar(wt)
        σ2new = similar(σ2)
        gradz = zeros(T, 2)
        infoz = zeros(T, 2, 2)
        Δz    = zeros(T, 2)
    end
    # initialize algorithm
    if V2 == I
        loglconst = - n * log(2π)
    else
        loglconst = - n * log(2π) - logdet(cholfact(Symmetric(V2)))
    end
    logl = (loglconst - sum(log, wt) - sum(storagen)) / 2
    verbose && println("iter=", 0," logl=", logl)
    verbose && println("β=", β," σ2=", σ2)
    niter = maxiter
    # MM/EM/FS loop
    for iter in 1:maxiter
        # update variance components
        if algo == :MM || algo == :EM
            # update σ2[2]
            storagen .= r̃ ./ wt
            qf = vecnorm(storagen)^2
            tr = sum(inv, wt)
            if algo == :MM
                σ2[2] *= sqrt(qf / tr)
            elseif algo == :EM
                σ2[2] += (qf - tr) / n * σ2[2]^2
            end
            # update σ2[1]
            storagen .= (r̃ ./ wt).^2 .* d
            qf = sum(storagen)
            storagen .= d ./ wt
            tr = sum(storagen)
            if algo == :MM
                σ2[1] *= sqrt(qf / tr)
            elseif algo == :EM
                σ2[1] += (qf - tr) / rk1 * σ2[1]^2
            end
            # update Ω
            wt .= σ2[1] .* d .+ σ2[2]
        elseif algo == :FS
            # update gradz
            storagen .= r̃ ./ wt
            qf = vecnorm(storagen)^2
            tr = sum(inv, wt)
            gradz[2] = (qf - tr) / 2 * σ2[2]
            storagen .= storagen.^2 .* d
            qf = sum(storagen)
            storagen .= d ./ wt
            tr = sum(storagen)
            gradz[1] = (qf - tr) / 2 * σ2[1]
            # update infoz
            infoz[1, 1] = σ2[1] * σ2[1] * sum(abs2, storagen) / 2
            storagen ./= wt
            infoz[1, 2] = infoz[2, 1] = σ2[1] * σ2[2] * sum(storagen) / 2
            storagen .= inv.(wt)
            infoz[2, 2] = σ2[2] * σ2[2] * sum(abs2, storagen) / 2
            # make sure the information matrix is positive definite
            if eigmin(infoz) ≤ 1e-3; infoz += T(1e-3)I; end
            # line search
            A_ldiv_B!(Δz, cholfact!(infoz), gradz) # Newton direction
            stepsize = T(1)
            for btiter in 1:50
                znew  .= z .+ stepsize .* Δz
                σ2new .= exp.(znew)
                wtnew .= σ2new[1] .* d .+ σ2new[2]
                if minimum(wtnew) < 1e-8
                    stepsize /= 2
                    continue
                end
                storagen .= r̃ .* r̃ ./ wtnew
                loglnew = (loglconst - sum(log, wtnew) - sum(storagen)) / 2
                if loglnew > logl
                    break
                elseif btiter == 50
                    warn("line search failed")
                else
                    stepsize /= 2
                end             
            end 
            copy!(z, znew)
            copy!(σ2, σ2new)
            copy!(wt, wtnew)
        end
        # update fixed effects
        storagen .= inv.(wt)
        copy!(storagenp, X̃)
        scale!(storagen, storagenp)
        At_mul_B!(storagepp, X̃, storagenp) # storagepp is the Gram matrix
        storagen .= storagen .* ỹ
        At_mul_B!(storagep, X̃, storagen)
        copy!(β, storagep)
        if !isposdef(storagepp)
            storagepp += T(1e-3)I
        end
        A_ldiv_B!(cholfact!(Symmetric(storagepp)), β)
        A_mul_B!(r̃, X̃, β)
        r̃ .= ỹ .- r̃
        storagen .= r̃ .* r̃ ./ wt
        # check convergence
        loglold = logl
        logl = (loglconst - sum(log, wt) - sum(storagen)) / 2
        verbose && println("iter=", iter," logl=", logl)
        verbose && println("β=", β," σ2=", σ2)
        if abs(logl - loglold) < funtol * (abs(logl) + 1)
            niter = iter
            break
        end
    end
    # output
    return β, σ2, niter, logl
end
